from functools import partial
from typing import Collection, Union

import numpy as np
import torch
from tqdm import tqdm

from simtorch.model.sim_model import SimilarityModel
from simtorch.similarity.base_similarity import BaseSimilarity


class Procrustes(BaseSimilarity):
    def __init__(
        self,
        sim_model1: SimilarityModel,
        sim_model2: SimilarityModel,
        device: Union[str, torch.cuda.device] = "cpu",
    ):
        self.similarity_name = "Procrustes"
        self.device = "cpu"  # FIXME: procrustes OOMs on GPU
        self.sim_model1 = sim_model1
        self.sim_model2 = sim_model2

        self.sim_model1.model.to(self.device)
        self.sim_model2.model.to(self.device)

    def procrustes_dist(self, X: torch.tensor, Y: torch.tensor):
        frobenius_norm = partial(torch.linalg.norm, ord="fro")
        nuclear_norm = partial(torch.linalg.norm, ord="nuc")

        X = self._normalize(X)
        Y = self._normalize(Y)

        X = X / frobenius_norm(X)
        Y = Y / frobenius_norm(Y)

        return 1 - nuclear_norm(X.t() @ Y).cpu()

    def compute(self, dataloader: Collection):
        dist_matrices = []
        for X, *_ in tqdm(dataloader, total=len(dataloader)):
            X = X.to(self.device)
            N = X.shape[0]

            _ = self.sim_model1.model(X)
            _ = self.sim_model2.model(X)

            batch_dist_matrix = np.zeros((self.sim_model1.n_layers, self.sim_model2.n_layers))

            # iterate through layers
            for i, (_, activation1) in enumerate(self.sim_model1.model_activations.items()):
                for j, (_, activation2) in enumerate(self.sim_model2.model_activations.items()):
                    activation1 = activation1.reshape(N, -1)
                    activation2 = activation2.reshape(N, -1)
                    dist = self.procrustes_dist(X=activation1, Y=activation2)
                    batch_dist_matrix[i, j] = dist.item()

            dist_matrices.append(batch_dist_matrix)

        self.sim_matrix = np.zeros_like(batch_dist_matrix)
        for mat in dist_matrices:
            self.sim_matrix += mat
        self.sim_matrix /= len(batch_dist_matrix)

        return self.sim_matrix



from typing import Collection, Union

import numpy as np
import torch
from tqdm import tqdm

from simtorch.model.sim_model import SimilarityModel
from simtorch.similarity.base_similarity import BaseSimilarity


class CKA(BaseSimilarity):
    def __init__(
        self,
        sim_model1: SimilarityModel,
        sim_model2: SimilarityModel,
        device: Union[str, torch.cuda.device] = "cpu",
        unbiased: bool = False,
    ):
        self.similarity_name = "CKA"
        self.device = device
        self.sim_model1 = sim_model1
        self.sim_model2 = sim_model2
        self.unbiased = unbiased

        self.sim_model1.model.to(self.device)
        self.sim_model2.model.to(self.device)

    def _centering(self, K: torch.tensor):
        n = K.shape[0]
        unit = torch.ones([n, n], device=self.device)
        identity = torch.eye(n, device=self.device)
        H = identity - unit / n
        return torch.matmul(torch.matmul(H, K), H)

    def linear_HSIC(self, L_X: torch.tensor, L_Y: torch.tensor):
        return torch.sum(self._centering(L_X) * self._centering(L_Y))

    def linear_CKA(self, L_X: torch.tensor, L_Y: torch.tensor):
        hsic = self.linear_HSIC(L_X, L_Y)
        var1 = torch.sqrt(self.linear_HSIC(L_X, L_X))
        var2 = torch.sqrt(self.linear_HSIC(L_Y, L_Y))

        return (hsic / (var1 * var2)).detach().cpu()

    def compute(self, dataloader: Collection):
        cka_matrices = []
        for X, *_ in tqdm(dataloader, total=len(dataloader)):
            X = X.to(self.device)
            batch_size = X.shape[0]

            # forward passes to activate hooks
            _ = self.sim_model1.model(X)
            _ = self.sim_model2.model(X)

            batch_cka_matrix = np.zeros((self.sim_model1.n_layers, self.sim_model2.n_layers))

            # iterate through layers
            for i, (_, activation1) in enumerate(self.sim_model1.model_activations.items()):
                X = self._normalize(activation1.view(batch_size, -1).to(self.device))
                L_X = torch.matmul(X, X.T)

                for j, (_, activation2) in enumerate(self.sim_model2.model_activations.items()):
                    Y = self._normalize(activation2.view(batch_size, -1).to(self.device))
                    L_Y = torch.matmul(Y, Y.T)

                    layer_cka = self.linear_CKA(L_X=L_X, L_Y=L_Y)

                    batch_cka_matrix[i, j] = layer_cka.item()

            cka_matrices.append(batch_cka_matrix)

        self.sim_matrix = np.zeros_like(batch_cka_matrix)
        for mat in cka_matrices:
            self.sim_matrix += mat
        self.sim_matrix /= len(batch_cka_matrix)

        return self.sim_matrix



from typing import Collection, Union

import numpy as np
import torch
from tqdm import tqdm

from simtorch.model.sim_model import SimilarityModel
from simtorch.similarity.base_similarity import BaseSimilarity
from simtorch.similarity.utils import remove_negative_eigenvalues


class DeltaCKA(BaseSimilarity):
    def __init__(
        self,
        sim_model1: SimilarityModel,
        sim_model2: SimilarityModel,
        device: Union[str, torch.cuda.device] = "cpu",
        remove_negative_eig: bool = False
    ):
        self.similarity_name = "dCKA"
        self.device = device
        self.sim_model1 = sim_model1
        self.sim_model2 = sim_model2
        self.remove_negative_eig = remove_negative_eig

        self.sim_model1.model.to(self.device)
        self.sim_model2.model.to(self.device)

    def _centering(self, K: torch.tensor):
        n = K.shape[0]
        unit = torch.ones([n, n], device=self.device)
        identity = torch.eye(n, device=self.device)
        H = identity - unit / n

        return torch.matmul(torch.matmul(H, K), H)

    def _linear_residual(self, input_y, input_dist):
        input_y = input_y.view(-1, 1)

        inv_term = torch.inverse(input_dist.T @ input_dist)
        beta = inv_term @ input_dist.T @ input_y
        y_pred = input_dist @ beta

        res_sim = (input_y - y_pred).view(-1)
        pve = (1 - res_sim.var() / input_y.var()).item()

        return res_sim, pve

    def linear_HSIC(
        self,
        L_X: torch.tensor,
        L_Y: torch.tensor,
        input_confounders: torch.tensor,
    ):
        input_grammian = (
            torch.matmul(input_confounders, input_confounders.T).view(-1, 1)
        )
        input_distance_1 = torch.cat(
            (torch.ones(input_grammian.shape[0], 1, device=self.device), input_grammian), 1
        )

        input_distance_2 = torch.cat(
            (torch.ones(input_grammian.shape[0], 1, device=self.device), input_grammian), 1
        )

        residuals1, pve_1 = self._linear_residual(L_X, input_distance_1)
        residuals2, pve_2 = self._linear_residual(L_Y, input_distance_2)

        residuals1 = residuals1.view(L_X.shape)
        residuals2 = residuals2.view(L_Y.shape)
        if self.remove_negative_eig:
            residuals1 = remove_negative_eigenvalues(residuals1)
            residuals2 = remove_negative_eigenvalues(residuals2)

        return (torch.sum(self._centering(residuals1) * self._centering(residuals2)), pve_1, pve_2)

    def linear_CKA(
        self,
        L_X: torch.tensor,
        L_Y: torch.tensor,
        input_confounders: torch.tensor,
    ):
        input_confounders = self._normalize(
            input_confounders.to(self.device))

        hsic, pve_1, pve_2 = self.linear_HSIC(L_X, L_Y, input_confounders)
        var1 = torch.sqrt(self.linear_HSIC(L_X, L_X, input_confounders)[0])
        var2 = torch.sqrt(self.linear_HSIC(L_Y, L_Y, input_confounders)[0])

        return (hsic / (var1 * var2)).detach().cpu(), pve_1, pve_2

    def compute(self, dataloader: Collection):
        cka_matrices = []
        for X, *_ in tqdm(dataloader, total=len(dataloader)):
            X = X.to(self.device)
            batch_size = X.shape[0]

            _ = self.sim_model1.model(X)
            _ = self.sim_model2.model(X)

            input_confounders = X.view(batch_size, -1)
            batch_cka_matrix = np.zeros((self.sim_model1.n_layers, self.sim_model2.n_layers))

            # iterate through layers
            for i, (_, activation1) in enumerate(self.sim_model1.model_activations.items()):
                X = self._normalize(activation1.view(batch_size, -1).to(self.device))
                L_X = torch.matmul(X, X.T)

                for j, (_, activation2) in enumerate(self.sim_model2.model_activations.items()):
                    Y = self._normalize(activation2.view(batch_size, -1).to(self.device))
                    L_Y = torch.matmul(Y, Y.T)

                    layer_cka, _, _ = self.linear_CKA(
                        L_X=L_X,
                        L_Y=L_Y,
                        input_confounders=input_confounders,
                    )
                    batch_cka_matrix[i, j] = layer_cka.item()

            cka_matrices.append(batch_cka_matrix)

        self.sim_matrix = np.zeros_like(batch_cka_matrix)
        for mat in cka_matrices:
            self.sim_matrix += mat
        self.sim_matrix /= len(batch_cka_matrix)

        return self.sim_matrix



"""
Thin wrapper over `torch.nn.Module()` to add forwards hooks and store
activations as attributes
"""
from functools import partial
from typing import Dict, List, Union
from warnings import warn

import torch


class SimilarityModel:
    def __init__(
        self,
        model: torch.nn.Module,
        model_name: Union[None, str] = None,
        layers_to_include: Union[None, List[str]] = None,
        device: Union[torch.cuda.device, str] = "cpu",
    ):
        self.model = model
        self.device = device
        self.layers_to_include = [] if layers_to_include is None else layers_to_include
        if not self.layers_to_include:
            warn("No layers have been added for similarity metric computation.")

        self.model.to(self.device)
        self.model.eval()
        self.model_name: str = model_name if model_name is not None else self.model.__class__.__name__
        self.model_activations: Dict[str, torch.tensor] = {}

        self.hook_model()

    def hook_model(self):
        for name, layer in self.model.named_modules():
            if any([x in name for x in self.layers_to_include]):
                layer.register_forward_hook(partial(self._add_hook, name))

    def _add_hook(self, name, layer, input, output):
        self.model_activations[name] = output

    @property
    def n_layers(self):
        if len(self.model_activations) == 0:
            warn(
                message="Model activations are empty, there's a chance `model.forward()` hasn't been run yet."
            )
        return len(self.model_activations.keys())



import os
from pathlib import Path
from typing import Collection, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch


class BaseSimilarity:
    def __init__(self, sim_model1, sim_model2, device="cpu", **kwargs):
        self.similarity_name = "base"

        self.sim_model1 = sim_model1
        self.sim_model2 = sim_model2
        self.device = device

        self.sim_matrix = np.zeros((3, 3))

    def _normalize(self, X):
        X = X - X.mean(0, keepdims=True)
        return X / torch.linalg.norm(X)

    def compute(self, dataloader: Collection):
        pass

    def plot_similarity(
        self,
        title: Union[None, str] = None,
        xlabel: Union[None, str] = None,
        ylabel: Union[None, str] = None,
        savefig: bool = False,
        save_path: Union[str, Path] = "../output",
        plot_cmap: str = "magma"
    ):
        sns.set(rc={"figure.dpi": 200, 'savefig.dpi': 200})
        title = (
            f"{self.similarity_name}: "
            f"{self.sim_model1.model_name} vs {self.sim_model2.model_name}"
            if title is None
            else title
        )
        xlabel = f"{self.sim_model2.model_name}" if xlabel is None else xlabel
        ylabel = f"{self.sim_model1.model_name}" if ylabel is None else ylabel

        ax = sns.heatmap(self.sim_matrix, cmap=plot_cmap, cbar=True)
        ax.set(xlabel=xlabel, ylabel=ylabel, title=title)
        if savefig:
            os.makedirs(save_path, exist_ok=True)
            filename = "_".join(title.split())
            plt.savefig(os.path.join(save_path, filename))

        return ax
